(* 配列の大域的定数畳み込み *)

open Global
open ILang

let remove env x = match x with Label x -> S2.remove x env | _ -> env

let rec g_store enable env e =
  match e with
  | V x -> remove env x, e
  | Op(Label ".<-", [Label x; Int i; z]) ->
      if enable && S2.mem x env && is_const z then begin
        (Hashtbl.find gl_vars x).(i) <- z;
        env, V Unit
      end else env, e
  | Op(Label x, _) when List.exists (fun y -> x = y.name) !gl_funs ->
      S2.empty, e
  | Op(_, xs) -> List.fold_left remove env xs, e
  | If(_,_,_,e1,e2) ->
      let env1, _ = g_store false env e1 in
      let env2, _ = g_store false env e2 in
      S2.inter env1 env2, e
  | Let(x,e1,e2) ->
      let env1, e1_ = g_store enable env e1 in
      if S2.is_empty env1 then env1, Let(x, e1_, e2) else
      let env2, e2_ = g_store enable env1 e2 in
      env2, Let(x,e1_,e2_)
  | LetRec _ -> assert false

let rec check env e =
  match e with
  | V x -> remove env x
  | Op(Label ".->", _) -> env
  | Op(_,xs) -> List.fold_left remove env xs
  | If(_,_,_,e1,e2) | Let(_,e1,e2) | LetRec(_,_,e1,e2) ->
      check (check env e1) e2

let rec g_load env e =
  match e with
  | Op(Label ".->", [Label x; Int i]) when S2.mem x env ->
      V (Hashtbl.find gl_vars x).(i)
  | V _ | Op _ -> e
  | If(c,x,y,e1,e2) -> If(c,x,y, g_load env e1, g_load env e2)
  | Let(x,e1,e2) -> Let(x, g_load env e1, g_load env e2)
  | LetRec(x,ys,e1,e2) -> LetRec(x,ys, g_load env e1, g_load env e2)

let f e =
  let s = Hashtbl.fold (fun k _ s -> S2.add k s) gl_vars S2.empty in
  let env = Hashtbl.fold (fun _ v s -> Array.fold_left remove s v) gl_vars s in
  let e1 = snd (g_store true env e) in
  let env1 = List.fold_left
      (fun env { body = e } -> check env e)
      (check env e) !gl_funs
  in
  gl_funs := List.map (fun x -> { x with body = g_load env1 x.body }) !gl_funs;
  g_load env1 e1
